require "import"
import "android.app.*"
import "android.os.*"
import "android.widget.*"
import "android.view.*"
import "android.media.*"
import "android.content.*"
import "android.Manifest"
import "android.content.pm.PackageManager"
import "java.io.*"
import "android.net.Uri"

activity.setTitle("Advanced Karaoke Recorder")

-- Permissions
function checkPermissions()
  if Build.VERSION.SDK_INT >= 23 then
    local permissions = {
      Manifest.permission.RECORD_AUDIO,
      Manifest.permission.WRITE_EXTERNAL_STORAGE,
      Manifest.permission.READ_EXTERNAL_STORAGE
    }
    for _, perm in ipairs(permissions) do
      if activity.checkSelfPermission(perm) ~= PackageManager.PERMISSION_GRANTED then
        activity.requestPermissions(permissions, 1)
        return
      end
    end
  end
end

checkPermissions()

-- Paths
recordDir = "/storage/emulated/0/Download/sujan/recording/"
File(recordDir).mkdirs()
tempFilePath = activity.getExternalFilesDir(nil).toString().."/temp_record.3gp"
karaokePath = nil
karaokeLoopMode = "once"

-- UI Layout
layout = {
  ScrollView,
  {
    LinearLayout,
    orientation="vertical",
    padding="16dp",
    {
      Button,
      text="About",
      onClick=function()
        AlertDialog.Builder(activity)
          .setTitle("About This App")
          .setMessage("Hi, this voice recorder is developed by Sujan Rai.\n\nPlease join my Telegram channel for apps, tutorials, and resources.")
          .setPositiveButton("Join Telegram", function()
            local intent = Intent(Intent.ACTION_VIEW, Uri.parse("https://t.me/techvisionarynepal"))
            activity.startActivity(intent)
          end)
          .setNegativeButton("OK", nil)
          .show()
      end
    },
    {
      Button,
      text="Help",
      onClick=function()
        AlertDialog.Builder(activity)
        .setTitle("How to Use")
        .setMessage([[
1. Tap 'Karaoke Settings' to upload a karaoke track and set loop mode.
2. Tap 'Start Recording' to begin recording your voice (karaoke will play if uploaded).
3. Tap again to stop recording.
4. Tap 'Play Recording' to preview.
5. Enter filename and tap 'Save Recording' to export as WAV.
6. Use rewind/forward buttons to jump playback.
        ]])
        .setPositiveButton("OK", nil)
        .show()
      end
    },
    {
      Button,
      id="settingsBtn",
      text="Karaoke Settings"
    },
    {
      Button,
      id="recordBtn",
      text="Start Recording"
    },
    {
      Button,
      id="playBtn",
      text="Play Recording",
      enabled=false,
      visibility="gone"
    },
    {
      LinearLayout,
      orientation="horizontal",
      visibility="gone",
      id="jumpControl",
      {
        Button,
        id="rewindBtn",
        text="<< 10s",
        layout_weight="1"
      },
      {
        Button,
        id="forwardBtn",
        text="10s >>",
        layout_weight="1"
      }
    },
    {
      EditText,
      id="fileNameInput",
      hint="Enter file name",
      visibility="gone"
    },
    {
      Button,
      id="saveBtn",
      text="Save Recording",
      enabled=false,
      visibility="gone"
    },
    {
      Button,
      id="resetBtn",
      text="Reset",
      layout_marginTop="8dp"
    },
    {
      Button,
      id="exitBtn",
      text="Exit",
      layout_marginTop="8dp"
    },
    {
      TextView,
      id="statusText",
      text="Status: Idle",
      textSize="16sp",
      layout_marginTop="16dp"
    }
  }
}

activity.setContentView(layout)

-- Vars
mediaRecorder = nil
mediaPlayer = nil
karaokePlayer = nil
isRecording = false
isPlaying = false

function startRecording()
  pcall(function()
    mediaRecorder = MediaRecorder()
    mediaRecorder.setAudioSource(MediaRecorder.AudioSource.MIC)
    mediaRecorder.setOutputFormat(MediaRecorder.OutputFormat.THREE_GPP)
    mediaRecorder.setAudioEncoder(MediaRecorder.AudioEncoder.AAC)
    mediaRecorder.setAudioEncodingBitRate(128000)
    mediaRecorder.setAudioSamplingRate(44100)
    mediaRecorder.setOutputFile(tempFilePath)
    mediaRecorder.prepare()
    mediaRecorder.start()

    if karaokePath then startKaraoke() end

    isRecording = true
    recordBtn.setText("Stop Recording")
    statusText.setText("Status: Recording...")
  end, function(err)
    statusText.setText("Error: "..err)
  end)
end

function stopRecording()
  pcall(function()
    mediaRecorder.stop()
    mediaRecorder.release()
    mediaRecorder = nil
    isRecording = false

    recordBtn.setText("Start Recording")
    playBtn.setVisibility(View.VISIBLE)
    playBtn.setEnabled(true)
    fileNameInput.setVisibility(View.VISIBLE)
    saveBtn.setVisibility(View.VISIBLE)
    saveBtn.setEnabled(true)
    jumpControl.setVisibility(View.VISIBLE)

    statusText.setText("Recording finished")

    if karaokePlayer then
      karaokePlayer.stop()
      karaokePlayer.release()
      karaokePlayer = nil
    end
  end, function(err)
    statusText.setText("Stop error: "..err)
  end)
end

function toggleRecording()
  if not isRecording then startRecording() else stopRecording() end
end

function togglePlayback()
  if isPlaying then stopPlayback() else playRecording() end
end

function playRecording()
  pcall(function()
    mediaPlayer = MediaPlayer()
    mediaPlayer.setDataSource(tempFilePath)
    mediaPlayer.prepare()
    mediaPlayer.start()
    isPlaying = true
    playBtn.setText("Stop Playback")
    statusText.setText("Playing...")

    mediaPlayer.setOnCompletionListener{
      onCompletion=function(mp)
        mp.release()
        mediaPlayer = nil
        isPlaying = false
        playBtn.setText("Play Recording")
        statusText.setText("Playback finished")
      end
    }
  end, function(err)
    statusText.setText("Playback error: "..err)
  end)
end

function stopPlayback()
  if mediaPlayer then
    pcall(function()
      mediaPlayer.stop()
      mediaPlayer.release()
      mediaPlayer = nil
      isPlaying = false
      playBtn.setText("Play Recording")
      statusText.setText("Playback stopped")
    end)
  end
end

function saveRecording()
  local name = fileNameInput.getText().toString()
  if name == "" then
    statusText.setText("Please enter file name")
    return
  end
  local outputFile = File(recordDir..name..".wav")
  pcall(function()
    local fis = FileInputStream(tempFilePath)
    local fos = FileOutputStream(outputFile)
    local buffer = byte[1024]
    local len
    while true do
      len = fis.read(buffer)
      if len == -1 then break end
      fos.write(buffer, 0, len)
    end
    fis.close()
    fos.close()
    statusText.setText("Saved as: "..outputFile.toString())
    saveBtn.setEnabled(false)
  end, function(err)
    statusText.setText("Save error: "..err)
  end)
end

function openKaraokeSettings()
  local loopOptions = {"once", "twice", "loop"}
  local selected = 0
  for i, v in ipairs(loopOptions) do
    if v == karaokeLoopMode then selected = i-1 break end
  end

  local builder = AlertDialog.Builder(activity)
  builder.setTitle("Karaoke Settings")
  builder.setSingleChoiceItems(loopOptions, selected, function(dialog, which)
    karaokeLoopMode = loopOptions[which+1]
  end)
  builder.setPositiveButton("Upload Karaoke", function()
    local intent = Intent(Intent.ACTION_GET_CONTENT)
    intent.setType("audio/*")
    activity.startActivityForResult(intent, 1001)
  end)
  builder.setNegativeButton("OK", nil)
  builder.show()
end

function startKaraoke()
  pcall(function()
    local uri = Uri.parse(karaokePath)
    karaokePlayer = MediaPlayer()
    karaokePlayer.setDataSource(activity, uri)
    karaokePlayer.setLooping(karaokeLoopMode == "loop")
    karaokePlayer.prepare()
    karaokePlayer.start()

    karaokePlayer.setOnCompletionListener{
      onCompletion=function(mp)
        if karaokeLoopMode == "twice" then
          karaokePlayer.start()
          karaokeLoopMode = "once"
        end
      end
    }
  end, function(err)
    statusText.setText("Karaoke error: "..err)
  end)
end

function onActivityResult(req, res, data)
  if req == 1001 and res == Activity.RESULT_OK then
    local uri = data.getData()
    karaokePath = uri.toString()
    statusText.setText("Karaoke loaded")
  end
end

-- Reset App State
function resetAll()
  AlertDialog.Builder(activity)
    .setTitle("Reset All")
    .setMessage("Are you sure you want to reset everything?")
    .setPositiveButton("Yes", function()
      stopRecording()
      stopPlayback()
      fileNameInput.setText("")
      karaokePath = nil
      karaokeLoopMode = "once"
      jumpControl.setVisibility(View.GONE)
      playBtn.setVisibility(View.GONE)
      fileNameInput.setVisibility(View.GONE)
      saveBtn.setVisibility(View.GONE)
      saveBtn.setEnabled(false)
      recordBtn.setText("Start Recording")
      statusText.setText("Status: Idle")
    end)
    .setNegativeButton("Cancel", nil)
    .show()
end

-- Playback Jump Buttons
rewindBtn.onClick = function()
  if mediaPlayer then
    local newPos = math.max(0, mediaPlayer.getCurrentPosition() - 10000)
    mediaPlayer.seekTo(newPos)
  end
end

forwardBtn.onClick = function()
  if mediaPlayer then
    local maxDur = mediaPlayer.getDuration()
    local newPos = math.min(maxDur, mediaPlayer.getCurrentPosition() + 10000)
    mediaPlayer.seekTo(newPos)
  end
end

-- Event Bindings
recordBtn.onClick = toggleRecording
playBtn.onClick = togglePlayback
saveBtn.onClick = saveRecording
settingsBtn.onClick = openKaraokeSettings
resetBtn.onClick = resetAll
exitBtn.onClick = function()
  AlertDialog.Builder(activity)
    .setTitle("Attention:")
    .setMessage("Do you really want to exit the tool?")
    .setPositiveButton("Yes", function() activity.finish() end)
    .setNegativeButton("No", nil)
    .show()
end

-- Intro Dialog
AlertDialog.Builder(activity)
  .setTitle("Voice Recorder with Karaoke:")
  .setMessage("Hi, this tool is developed by Sujan Rai at Tech Visionary.")
  .setPositiveButton("OK", nil)
  .show()